package require Itcl
package require Itk
namespace import -force itcl::*
namespace import -force itk::*
lappend auto_path $env(ANTELOPE)/data/tcl/library
package require Dbe

class MapWidget {

    inherit itk::Widget

    itk_option define -buttons buttons Buttons {} {}

    constructor {args} {
	set w $itk_interior.map

	global Graticule
	set Graticule($this) $graticule
	global Projection
	set Projection($this) $projection

        # itk_option add MapWidget::buttons
# the geometry just constrains the minimum size of the map, 
# so we want it small.
	itk_component add graphic { 
	    graphic $w \
		-command "$this _refresh" \
		-geometry $geometry \
		-doublebuffered yes \
		-squareaspect no
	} {
	    keep -background
	}

	itk_component add buttons { 
	    frame $itk_interior.f
	}

	bind $w <ButtonPress-3>  "$this _show %X %Y %x %y"
	bind $w <B1-Motion>  "$this _scan $w %x %y"
	bind $w <B1-ButtonRelease> "$this _zoom $w"
	bind $w <B2-Motion> "$this _scroll %x %y"
	bind $w <B2-ButtonRelease> "$this _redraw_scrolled"

	pack $w -side top -fill both -expand yes
	if { $label != "" } { 
	    set wlabel $itk_interior.label
	    itk_component add label { 
		label $wlabel 
	    }
	    pack $wlabel -side top -fill x
	}

	pack $itk_component(buttons) -side top -fill x

	eval itk_initialize $args
	add_buttons 

	eval pj_init -c $current_center -r $current_range -p $projection
	set repaint 1

	set built 1
    }

    method widget {} { return $w } 

    proc Pj_inv {x y}  {
	set return [pj_inv $x $y]
	if { [regexp Inf $return] } {
	    bgerror "pj_inv returns $return for $x $y"
	}
	return $return
    }

    method postscript {file} {
	set save $current_hires
	if { $current_range < $postscript_hires } { 
	    set current_hires 1
	}
	g_output $file
	g_newdev cps
	_paintme 
	g_exit
	set current_hires $save
    }

    method save {} {
	set filename [tk_getSaveFile \
		-initialdir . \
		-title "Save postscript map" \
		-defaultextension .ps \
		-parent $itk_interior]
	if { $filename == "" } {
	    return 
	}
	busy hold $itk_interior -cursor watch
	update
	if { [catch "postscript $filename" result] } { 
	    bgerror $result
	}
	busy release $itk_interior
    }

    method print {} {
	set tmp /tmp/##amap[pid].ps
	postscript $tmp
	exec lp $tmp 
	after 500 exec rm $tmp
    }

    protected method add_buttons {} { 
	foreach i $existing_buttons { 
	    destroy $i
	}
	foreach i $itk_option(-buttons) {
	    switch $i {

		coverage {
		    set b $itk_component(buttons).coverage
		    itk_component add coverage { 
			menubutton $b \
			    -text Coverage \
			    -relief raised \
			    -menu $b.menu
		    }
		    itk_component add coverage_menu { 
			menu $b.menu -tearoff no
		    }
		    foreach i { coasts rivers political states counties } {
			$b.menu add checkbutton \
			    -label $i \
			    -variable Coverage($i$this) \
			    -command "$this set_coverage"
		    }
		    global Coverage
		    foreach i $coverage {
			set Coverage($i$this) 1
		    }
		}

		graticule {
		    global Graticule
		    set b $itk_component(buttons).graticule
		    itk_component add graticule { 
			checkbutton $b \
			    -text Graticule \
			    -variable Graticule($this) \
			    -command "$this redraw"
		    }
		    set Graticule($this) $graticule
		}

		hires	{
		    set b $itk_component(buttons).hires
		    itk_component add hires { 
			button $b \
			    -text HiRes \
			    -command "$this hires"
		    }
		}

		print {
		    set b $itk_component(buttons).print
		    itk_component add print { 
			button $b \
			    -text Print \
			    -command "$this print"
		    }
		}

		projection	{ 
		    global Projection
		    set b $itk_component(buttons).projection
		    set old $projection
		    set Projection($this) 
		    itk_component add projection { 
		 	SelectorButton $b \
		 	    -options "aeqd robin merc mill moll vandg" \
			    -variable Projection($this) \
		 	    -command "$this set_projection"
		    }
		    $b choose $old
		}

		redraw	{ 
		    set b $itk_component(buttons).redraw
		    itk_component add redraw { 
			button $b \
			    -text Redraw \
			    -command "$this redraw"
		    }
		}

		save {
		    set b $itk_component(buttons).save
		    itk_component add save { 
			button $b \
			    -text Save \
			    -command "$this save"
		    }
		}

		
		zoomout {
		    set b $itk_component(buttons).zoomout 
		    itk_component add zoomout { 
			button $b \
			    -text "Zoom out" \
			    -command "$this zoom_out"
		    }
		}

	    }
	    lappend existing_buttons $b
	    pack $b -side left
	}

	if { $quit } { 
	    set b $itk_component(buttons).quit
	    itk_component add quit { 
		button $b \
		    -text Quit \
		    -fg white \
		    -bg red -activebackground red \
		    -command "destroy ."
	    } {
		keep -font
	    }
	    pack $b -side right
	}

	if { $indicator } { 
	    global Indicator
	    set b $itk_component(buttons).indicator
	    set Indicator($this) 0
	    itk_component add indicator { 
		checkbutton $b \
		    -text Location \
		    -variable Indicator($this) \
		    -command "$this indicate"
	    }
	    pack $b -side left

	    set b $itk_component(buttons).entry
	    set location_indicator $b
	    itk_component add location_indicator { 
		entry $b \
		    -width 15 \
		    -relief sunken
	    } { 
		keep -background
	    }
	    bind $b <Return> "$this _recenter"
	    pack $b -side left
	}
    }

    method _recenter {} {
	set current_center [$location_indicator get]
	eval pj_init -c $current_center -r $current_range -p $projection
	redraw
    }

    method indicate {} {
	global Indicator
	if { $Indicator($this) } { 
	    bind $w <Motion> "$this showloc %X %Y %x %y"
	} else { 
	    bind $w <Motion> ""
	}
    }

    method showloc {X Y x y} {
	eval set a [g_toworld $x $y]
	set gx [format %10.2f [lindex $a 0]]
	set gy [format %10.2f [lindex $a 1]]
	set latlon [pj_inv $gy $gx]
	if { ! [regexp Inf $latlon] } {
	    set prec [max 0 [expr -int(floor(log10($current_range)-1))]]
	    set format "%6.${prec}f %6.${prec}f"
	    set latlon [eval format {$format} $latlon]
	}
	$location_indicator delete 0 end 
	$location_indicator insert 0 $latlon
    }

    method set_coverage {} {
	global Coverage
	set coverage ""
	foreach i [array names Coverage *$this] {
	    if { $Coverage($i) == 1 } {
		regsub $this $i {} acover
		lappend coverage $acover
	    }
	}
	$this redraw
    }

    method set_projection {proj} {
	global Projection
	set projection $proj
	set Projection($this) $projection
	eval pj_init -c $current_center -r $current_range -p $Projection($this)
	$this redraw
    }

    method hires {} {
	set current_hires 1
	$this redraw
	set current_hires 0
    }

    method zoom_out {} {
	set_current_range [min 180 [expr $zoom_factor*$current_range]]
	$this redraw
    }

    method redraw {args} {
	set repaint 1
	$itk_component(graphic) paint 
    }

    method _draw {} { 
	if { $current_hires } { 
	    eval drawmap -h $current_center $current_range \"$coverage\"
	} else { 
	    eval drawmap $current_center $current_range coasts
	}
    }

    method get {} {
	return "$current_center $current_range"
    }

    method _paintme {} {
	g_clear
	$this _draw

	global Graticule
	if { $Graticule($this) } { 
	    g_color gray
	    set aspect [g_getaspect]
	    if { $aspect < 1 } { 
		set lat_range $current_range
		set lon_range [expr $current_range/$aspect]
	    } else {
		set lon_range $current_range
		set lat_range [expr $current_range*$aspect]
	    }
	    eval ::graticule $current_center $lat_range $lon_range
	}

	if { $command != "" } { 
	    eval $command $current_center $current_range
	}
    }


    method _refresh {} {
	if { $built == 0 } { return } 
	set w [g_getdev]
	set ww [winfo width $w]
	set wh [winfo height $w]
	g_backbuffer
	if { $repaint || $wh != $height || $ww != $width } { 
	    set repaint 0
	    set height $wh
	    set width $ww

	    if { $current_range < 150 } { 
		set title "grname $current_center"
	    } else { 
		global Projection
		set title $Projection($this)
	    }
	    set title [ eval format {{%s (%6.3f, %7.3f) %7.3f}} {$title} $current_center $current_range]
	    if { $current_hires } { 
		append title " *"
	    }


	    if { $toplevel != "" } {
		wm title $toplevel "** $title **"
	    }
	    if { $wlabel != "" } {
		$wlabel config -text "$title" 
	    }

	    if { $toplevel != "" } {
		busy hold $toplevel -cursor watch
	    } else { 
		busy hold $itk_interior -cursor watch
	    }

	    update

	    if { [g_getdepth] > 1 } { 
		g_color blue
	    } else { 
		g_color white 
	    }

	    _paintme

	    if { $toplevel != "" } {
		wm title $toplevel "$title" 
	    }
	    if { $wlabel != "" } {
		$wlabel config 
	    }

	    if { $toplevel != "" } {
		busy release $toplevel 
	    } else { 
		busy release $itk_interior
	    }
	}
	g_swapbuffers
	g_frontbuffer
	$this _drawbox
    }

    method _drawbox {} {
	if { [info exists box(B,y)] } {
	    eval set a [g_toworld $box(A,x) $box(A,y)]
	    eval set b [g_toworld $box(B,x) $box(B,y)]
	    g_color red
	    g_polyfill 0
	    eval g_rect [lrange $a 0 1] [lrange $b 0 1]
	}
    }

    method _scan {w x y} { 
	if { [info exists box(A,x)] } {
	    if { abs($last(x,$w)-$x) < 5 && abs($last(y,$w)-$y) < 5 } { return }
	    set last(x,$w) $x
	    set last(y,$w) $y
	    set box(B,x) $x
	    set box(B,y) $y
	    $this _refresh
	} else { 
	    set last(x,$w) $x
	    set last(y,$w) $y
	    set box(A,x) $x
	    set box(A,y) $y
	}
    }

    method _show {X Y x y} {
	eval set a [g_toworld $x $y]
	set gx [format %10.2f [lindex $a 0]]
	set gy [format %10.2f [lindex $a 1]]
	set latlon [Pj_inv $gy $gx]
	if { $locate != "" } {
	    eval $locate $X $Y $x $y $latlon
	}
    }

    method xytomap {x y} {
	eval set a [g_toworld $x $y]
	set gx [format %10.2f [lindex $a 0]]
	set gy [format %10.2f [lindex $a 1]]
	set latlon [Pj_inv $gy $gx]
	return $latlon
    }

    method _zoom {w} {
	if { [info exists box(B,y)] } { 
	    set xc [expr ($box(A,x)+$box(B,x))/2]
	    set yc [expr ($box(A,y)+$box(B,y))/2]
	    eval set c [g_toworld $xc $yc]
	    set current_center [Pj_inv [lindex $c 1] [lindex $c 0]]

	    eval set a [g_toworld $box(A,x) $box(A,y)]
	    eval set b [g_toworld $box(B,x) $box(B,y)]
	    set latlonA [Pj_inv [lindex $a 1] [lindex $a 0]]
	    set latlonB [Pj_inv [lindex $b 1] [lindex $b 0]]
	    set dlat [expr abs([lindex $latlonA 0]-[lindex $latlonB 0])/2.0]
	    set dlon [expr abs([lindex $latlonA 1]-[lindex $latlonB 1])/2.0]
	    set latc [lindex $current_center 0]
	    set new [max $dlat [expr $dlon*cos($latc*57.2958)]]
	    if { $new < .001 } { return }
	    set_current_range $new
	} else {
	    set current_center $center
	    set_current_range $range
	}

	global Projection
	eval pj_init -c $current_center -r $current_range -p $Projection($this)
	set repaint 1
	catch { unset box(A,x) }
	catch { unset box(A,y) }
	catch { unset box(B,x) }
	catch { unset box(B,y) }
	$this _refresh
    }

    method set_current_range {range} {
	if { $range < .001 } { 
	    bgerror "range is $range [traceback]"
	    return 
	}
	set current_range [max $range .001]
	if { $current_range < $auto_hires } { 
	    set current_hires $hires
	} else { 
	    set current_hires 0
	}
	set stasz [expr $current_range/180. * 200000.]
	set orgsz [expr $stasz/2]
    }

    method _scroll {x y} { 
	if { [info exists scroll(x0)] } {
	    set scroll(x1) $x
	    set scroll(y1) $y
	    g_color white
	    g_scroll [expr $x-$scroll(x0)] [expr $y-$scroll(y0)]
	} else { 
	    set scroll(x0) $x
	    set scroll(y0) $y
	}
    }

    method _redraw_scrolled { } { 
	global Projection
	set x [expr int($width/2.0+$scroll(x0)-$scroll(x1)+.05)] 
	set y [expr int($height/2.0+$scroll(y0)-$scroll(y1)+.05)]
	eval set b [g_toworld $x $y]
	set current_center [Pj_inv [lindex $b 1] [lindex $b 0]]
	eval pj_init -c $current_center -r $current_range -p $Projection($this)
	set repaint 1
	catch [unset scroll]
	$this _refresh
    }

    method map2xy { lat lon } {
	set trans [pj_fwd $lat $lon]
	set y [lindex $trans 0]
	set x [lindex $trans 1]
	return  "$x $y"
    }


    public variable projection robin  { 
	global Projection
	set_projection $projection
    }

    public variable hires 1  
    public variable auto_hires 2
    public variable postscript_hires 30
    public variable graticule 0  {
	global Graticule
	set Graticule($this) $graticule
    }

# possible coverages are "fill coasts rivers political states counties"
    public variable coverage "fill coasts rivers political"

    public variable center {0 0} {
	set current_center $center
    }
    public variable range {180} {
	set_current_range $range
    }

    public variable geometry 100x100

    public variable locate ""
    public variable command ""
    public variable indicator 0
    public variable toplevel ""
    public variable label ""
    protected variable wlabel {}
    public variable quit 0
    public variable zoom_factor 5

    protected variable current_hires 0
    protected variable current_center {0 0} 
    protected variable current_range {180} 
    protected variable orgsz 1000 
    protected variable stasz 1000 
    protected variable w
    protected variable repaint 1
    protected variable height 0
    protected variable width 0
    protected variable scroll
    protected variable box
    protected variable last
    protected variable existing_buttons {}
    protected variable location_indicator
    protected variable built 0
}


# $Id$ 
